@@ -295,3 +295,12 @@ darcsPristineRefs fs = do
     where valid NoHash = False
           valid _ = True
 
+darcsCheckExpand :: Tree IO
+            -> IO (Either [(FilePath, Hash, Maybe Hash)] (Tree IO))
+darcsCheckExpand t = do
+  problemsOrTree <- checkExpand darcsHash t
+  case problemsOrTree of
+    Left problems -> return . Left $ map render problems
+    Right tree -> return . Right $ tree
+  where
+    render (path, h, h') = (anchorPath "." path, h, h')
\ No newline at end of file
